/*
 * @Date: 2022-07-01 16:03:32
 * @LastEditors: 刘君
 * @LastEditTime: 2022-07-05 10:04:23
 * @FilePath: \rpc-topic\src\game_post.ts
 */
import axios, { Axios } from "axios";
import { notification } from 'ant-design-vue'
import { useControlStore } from "/@/store/modules/control";
import { useUserStore } from "/@/store/modules/user";
export class GameClient {
    http: Axios
    user: string = '';
    password: string = '';
    constructor() {
        let baseURL = this.getUrl();
        const userStore = useUserStore();
        let userInfo = userStore.getUserInfo;
        this.http = new axios.Axios({ baseURL, headers: { 'Authorization': 'Basic ' + btoa(userInfo.user + ':' + userInfo.token) }, responseType: 'json' });
        this.http.interceptors.response.use(res => {
            if (res.status != 200) {
                notification.error({ message: "获取数据出错!", description: res.data });
                return Promise.reject()
            }
            try {
                res.data = JSON.parse(res.data);
                return res
            } catch (error) {
                notification.error({ message: "获取格式错误!", description: error as any });
                return Promise.reject()
            }
        });
    }
    setAuth(user, psw) {
        this.user = user;
        this.password = psw;
    }
    getUrl() {
        const store = useControlStore()
        let obj: any = store.getAllIP && store.getAllPort ? 'http://' + store.getAllIP + ':' + store.getAllPort + '/' : null;
        let baseURL = obj || 'http://127.0.0.1:8099/'
        return baseURL
    }
    query(url: string, data?: Record<string, any>, config?: Record<string, any>) {
        this.http.defaults.baseURL = this.getUrl();
        
        // let grafanaUser = (window as any).grafanaBootData.user
        // let user = grafanaUser.login + '_' + grafanaUser.id + "_" + grafanaUser.email
        let _data = {
            user: this.user || ' ',
            password: this.password || '',
            add_http_cors_header: 1,
            max_result_rows: 1000,
            max_result_bytes: 10000000,
            default_format: 'JSONCompact',
            result_overflow_mode: 'break',
        }
        if (data) {
            Object.assign(_data, data)
        }
        return _data
    }
    post(url, data?) {
        let _data = this.query(url, data)
        return this.http.post(`${url}`, JSON.stringify(_data))
    }
    get(url, data?) {
        let _data = this.query(url, data);
        return this.http.get(`${url}`, { params: _data })
    }
}

export const gameClient = new GameClient()