import { getPlugin } from "./plugins_mgr"
import { sendMsg } from "pi_pt/util/pid_rpc";
import { asyncSpawnReceive } from 'pi_pt/util/pid';
import { getExtVid } from "./gray_lib";

const timeoutFun = (time) => {
    return new Promise((resolve) => {
        setTimeout(() => {
            resolve("timeout");
        }, time);
    });
}

// 发送pid消息
export const sendRpcPid = async (name: string, msg: any, timeout: number = 3000, gray?: number) => {
    const vid = getExtVid(gray);
    if (!vid) {
        throw new Error("ext vid not found");
    }
    // const pid = Number(BigInt(Number(vid)) << BigInt(32));
    // console.log(`sendRpcPid pid ${pid}`);
    return sendRpcPidWithTimeout(async () => {
        // sendMsg(pid, name, msg)
        return await eventHandle(vid, name, msg);
    }, timeout);
}


// 封装promise超时
export const sendRpcPidWithTimeout = async (handle: () => void, timeout: number = 3000) => {
    const res = await Promise.any([handle(), timeoutFun(timeout)]);
    if (res == "timeout") {
        throw new Error("pid rpc timeout");
    }
    return res;
}


let pids: Map<string, number> = new Map();

// 指定虚拟机创建pid,并执行逻辑
const eventHandle = async (vid: number, event: string, msg: string) => {
    let key = `${vid}`;
    let pid = pids.get(key);
    if (!pid) {
        // 创建pid
        pid = await asyncSpawnReceive("", vid);
        pids.set(key, pid);
    }
    return await sendMsg(pid, event, msg);
}