import { Env } from "pi_ucore/db/env";
import { Log } from "pi_pt/util/log";
import { Select, DbItem, DataR, Data } from "management_page/server/rpc/db_mgr.struct";
import { BonBuffer } from "pi_serialization/bon";
import { checkAuth } from "./user.event";
import { DBType, Item } from "pi_ucore/db/mgr";
import { Tr } from "pi_pt/db/mgr";
import { HttpRequest, HttpResponse } from 'pi_pt/init/httprpc'
import { TabDecode, TabEncode } from "pi_serialization/tab_meta";
import { TabMeta } from "pi_serialization/sinfo";
import { WebShellMgr } from "pi_pt/native/pi_serv_lib/js_webshell";
import { checkHttpBasicAuth } from "./auth_lib";
import { getGrayV } from "pi_pt/util/util";
import { checkHttpPortAuth, getEnv } from "./http_proxy.event";
import { AUTH, FM, RESULT_NUM } from "../../../constant";

declare var ESProcess;
declare var env: Env;

/**
 * 创建webshell
 */
// #[httpRpc=/mgr/webshell/create]
export const create = async (req: HttpRequest, resp: HttpResponse) => {
    // 检测授权信息
    // checkHttpBasicAuth(req, resp);
    let userData = await checkHttpPortAuth(req);
    let localEnv = await getEnv(userData.user);
    if (!checkAuth(FM.WEB_SHELL, AUTH.W, localEnv)) throw "权限错误";
    let mgr = WebShellMgr.new(BigInt(getGrayV()));
    let pid = await mgr.crate_web_shell();
    return pid;
}


/**
 * 控制台输入
 * 代码补全 
    {
        "pid": 4294967299,
        "cmd": {
            "event": "tab",
            "msg": ""
        }
    }
 * 执行代码
    {
        "pid": 4294967299,
        "cmd": {
            "event": "input",
            "msg": "console.log(/"hello world/")"
        }
    }
 */
// #[httpRpc=/mgr/webshell/input]
export const input = async (req: HttpRequest, resp: HttpResponse) => {
    // 检测授权信息
    // checkHttpBasicAuth(req, resp);
    let userData = await checkHttpPortAuth(req);
    let localEnv = await getEnv(userData.user);
    if (!checkAuth(FM.WEB_SHELL, AUTH.W, localEnv)) throw "权限错误";
    let mgr = WebShellMgr.new(BigInt(getGrayV()));
    let data = JSON.parse(req.getData()[""]);
    let pid = Number(data.pid);
    let cmd = data.cmd;
    return mgr.client_input(pid, JSON.stringify(cmd));
}

/**
 * 获取输出
 * 定时获取，没有输出就返回空串
 */
// #[httpRpc=/mgr/webshell/out]
export const out = async (req: HttpRequest, resp: HttpResponse) => {
    // 检测授权信息
    // checkHttpBasicAuth(req, resp);
    let userData = await checkHttpPortAuth(req);
    let localEnv = await getEnv(userData.user);
    if (!checkAuth(FM.WEB_SHELL, AUTH.W, localEnv)) throw "权限错误";
    let mgr = WebShellMgr.new(BigInt(getGrayV()));
    let data = req.getData();
    let pid = Number(data.pid);
    return mgr.get_out(pid);
}

/**
 * 关闭控制台
 */
// #[httpRpc=/mgr/webshell/close]
export const close = async (req: HttpRequest, resp: HttpResponse) => {
    // 检测授权信息
    checkHttpBasicAuth(req, resp);
    let userData = await checkHttpPortAuth(req);
    let localEnv = await getEnv(userData.user);
    if (!checkAuth(FM.WEB_SHELL, AUTH.W, localEnv)) throw "权限错误";
    let data = req.getData();
    let pid = Number(data.pid);
    let mgr = WebShellMgr.new(BigInt(getGrayV()));
    return mgr.close(pid);
}

/**
 * 获取所有控制台
 */
// #[httpRpc=/mgr/webshell/all]
export const all = async (req: HttpRequest, resp: HttpResponse) => {
    // 检测授权信息
    // checkHttpBasicAuth(req, resp);
    let userData = await checkHttpPortAuth(req);
    let localEnv = await getEnv(userData.user);
    if (!checkAuth(FM.WEB_SHELL, AUTH.R, localEnv)) throw "权限错误";

    let mgr = WebShellMgr.new(BigInt(getGrayV()));
    return mgr.get_web_shell_all();
}